// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

#include "book.h"

// Heap class for a MAX heap (higher numbers at the top)
template <typename E>
class heap
{
    private:
        E *Heap;          // Pointer to the heap array
        int maxsize;      // Maximum size of the heap
        int n;            // Number of elements now in the heap

        // Helper function to put element in its correct place
        // Iterative version of siftdown
        void siftdown_iter(int pos)
        {
            while(!isLeaf(pos)) // Stop if pos is a leaf
            {
                int greaterChild = leftchild(pos);
                int rc = rightchild(pos);

                if((rc < n) && (Heap[rc] > Heap[greaterChild]))
                    greaterChild = rc;            // Set to greater child's value

                if(Heap[pos] > Heap[greaterChild])
                    return; // Done

                std::swap(Heap[pos], Heap[greaterChild]);

                pos = greaterChild;             // Move down (change index)
            }
        }

        // Recursive version of siftdown 
        void siftdown(int i)
        {
            int l = leftchild(i);
            int r = rightchild(i);
            int largest = i;

            if(l < n && Heap[l] > Heap[i])
                largest = l;

            if(r < n && Heap[r] > Heap[largest])
                largest = r;

            if(largest != i)
            {
                std::swap(Heap[i], Heap[largest]);
                siftdown(largest);
            }
        }


    public:
        // Constructor. Why no use of new?
        heap(E *h, int num, int max_init)     
        {
            Heap = h;
            n = num;
            maxsize = max_init;
            buildHeap();
        }

        int size() const       // Return current heap size
        {
            return n;
        }

        bool isLeaf(int pos) const // True if pos is a leaf
        {
            return ((pos >= n/2) && (pos < n));
        }

        // Indexing starting at 0
        int leftchild(int pos) const
        {
            return ((2 * pos) + 1);
        }    // Return leftchild position

        int rightchild(int pos) const
        {
            return ((2 * pos) + 2);
        }    // Return rightchild position

        int parent(int pos) const  // Return parent position
        {
            return ((pos - 1) / 2);
        }

        void buildHeap()           // Heapify contents of Heap
        {
            for(int i = parent(n - 1); i >= 0; i--)
            {
                //cout << "i is : " << i << endl;
                // Could be either siftdown method above
                siftdown(i);
            }
        }

        // Insert "it" into the heap
        void insert(const E &it)
        {
            Assert(n < maxsize, "Heap is full, write a grow function...");

            Heap[n] = it;            // Start after end of heap
            int curr = n++;          // ask, what does post increment do?

            // Now sift up until curr's parent > curr and not the root (0)
            while((curr != 0) && (Heap[curr] > Heap[parent(curr)]))
            {
                std::swap(Heap[curr], Heap[parent(curr)]);

                curr = parent(curr);
            }
        }

        // Remove first value
        E removefirst()
        {
            Assert(n > 0, "Heap is empty");

            // Swap first with last value
            std::swap(Heap[0], Heap[--n]);

            if(n != 0)
                // Could be either siftdown method above
                siftdown_iter(0);  // Siftdown new root val

            // Return "deleted" value (hack for temp storage after end of real data)
            return Heap[n];
        }
};

