/**
 * Heap algorithms
 * http://en.cppreference.com/w/cpp/algorithm
 * http://www.cplusplus.com/reference/algorithm/
 **/

#include <iostream>
#include <algorithm>
#include <vector>

using namespace std;

int main()
{
    vector<int> v{3, 1, 4, 1, 5, 9};

    cout << "initially, v: ";

    for(int i : v)
        cout << i << ' ';

    cout << endl;

    /**
      Checks if the elements in range [first, last) are a max heap.
      returns true if the range is max heap, false otherwise.
      Iterators for heap algorithms must meet the requirements of RandomAccessIterator.
     **/
    if(!is_heap(v.begin(), v.end()))
    {
        cout << "making heap..." << endl;

        /**
          Constructs a max heap in the range [first, last).
         **/
        make_heap(v.begin(), v.end());
    }

    cout << "after make_heap, v: ";

    for(int i : v)
        cout << i << ' ';

    cout << endl;

    v.push_back(6);
    cout << "before push_heap, v: ";
    
    for(int i : v)
        cout << i << ' ';

    cout << endl;

    /**
      Inserts the element at the position last-1 into the max heap defined by the range [first, last-1).
     **/
    push_heap(v.begin(), v.end());

    cout << "after push_heap, v: ";

    for(int i : v)
        cout << i << ' ';

    cout << endl << endl;

    /**
      Swaps the value in the position first and the value in the position last-1 and makes the subrange [first, last-1) into a max heap.
      This has the effect of removing the first (largest) element from the heap defined by the range [first, last).
     **/
    std::vector<int> v2 { 3, 1, 4, 1, 5, 9 };

    std::make_heap(v2.begin(), v2.end());

    std::cout << "v heapified: ";

    for(auto i : v2)
        std::cout << i << ' ';

    std::cout << '\n';

    std::pop_heap(v2.begin(), v2.end()); // moves the largest to the end

    std::cout << "after pop_heap: ";

    for(auto i : v2)
        std::cout << i << ' ';

    std::cout << '\n';

    int largest = v2.back();
    v2.pop_back();  // actually removes the largest element
    std::cout << "largest element: " << largest << '\n';

    std::cout << "heap without largest: ";

    for(auto i : v2)
        std::cout << i << ' ';

    std::cout << '\n' << endl;

    /**
      Converts the max heap [first, last) into a sorted range in ascending order.
      The resulting range no longer has the heap property.
     **/
    std::sort_heap(v2.begin(), v2.end());

    std::cout << "sorted:\t";
for (const auto &i : v2) {
        std::cout << i << ' ';
    }
}
