/**
 * Priority queue cointainer
 * http://en.cppreference.com/w/cpp/container/priority_queue
 * http://www.cplusplus.com/reference/queue/priority_queue/
 **/

#include <iostream>
#include <queue>
#include <vector>

using namespace std;


template<typename T>
void print_queue(T &q)
{
    while(!q.empty())
    {
        cout << q.top() << " ";
        q.pop();
    }

    cout << endl;
}

int main()
{
    priority_queue<int> qTemp;

    // defaults (same as above). Max-heap
    priority_queue<int, vector<int>, less<int> > qdefault;

    qdefault = qTemp;
    priority_queue<int> q(qTemp);

    qTemp.push(12);

    // no need to store the array
    for(auto n : {1, 8, 5, 6, 3, 4, 0, 9, 7, 2})
        q.push(n);

    cout << "Size is: " << q.size() << endl;

    if(!q.empty())
        print_queue(q);

    cout << "Size is: " << q.size() << endl;
    cout << endl;

    // Can construct from compatible container:
    vector<int> vec{3, 1, 4, 1, 5};
    priority_queue<int> q3(less<int>(), vec);

    cout << "Size is: " << q3.size() << endl;
    print_queue(q3);
    cout << "Size is: " << q3.size() << endl;
    cout << endl;

    // The greater version (min heap) requires template specificiation:
    priority_queue<int, vector<int>, greater<int> > q5(greater<int>(), vec);

    cout << "Size is: " << q5.size() << endl;
    print_queue(q5);
    cout << "Size is: " << q5.size() << endl;
    cout << endl;

    priority_queue<int, vector<int>, greater<int> > q2;

    for(auto n : {1, 8, 5, 6, 3, 4, 0, 9, 7, 2})
        q2.push(n);

    cout << "Size is: " << q2.size() << endl;

    priority_queue<int, vector<int>, greater<int> > qLast;
    qLast = q2;

    if(!qLast.empty())
        print_queue(qLast);

    cout << "Size is: " << qLast.size() << endl;
    cout << endl;

    return 0;
}
